#include <oxstd.h>
#include <oxdraw.h>
#include <oxprob.h>
#include <oxfloat.h>
#include <arma.h>
#import <modelbase>
#import <simula>
#import <maximize>
#import <database>
#import <maxsqp>

enum{GAUSS, STUD_T, SIGMA, LOG_SIGMA, INV_FISHER, INV_SQRT_FISHER, HESS, SAND};

static decl s_mY, s_cT, s_vf, s_vScore, s_vScaledSc, s_iSandw, s_vFunc;
static decl s_iLinkFunction, s_iScalingChoice, s_iDistribution, s_ip, s_iq, s_cPar, s_iStdErr;
			   
StartingValues(dOmega, const vA, const vB, const dMu, const dDf)
{
	decl asA, asB;
	if(s_iLinkFunction == SIGMA){ // In this case omega is estimated with exp(omega)
		if(dOmega <= 0) dOmega = -10;
		else dOmega = log(dOmega);
	}
	if(rows(vA) != s_ip || rows(vB) != s_iq){
		println("Number of starting values in vector vA, vB must be equal to s_ip, s_iq \nProgram stopped");
		exit(0);
	}
	if(sumc(vB) > 1){
		println("Sum of elements in vector vB cannot exceed 1 \nProgram stopped");
		exit(0);
	}
	asA = new array[1][s_ip];
	asB = new array[1][s_iq];
	for(decl i = 0; i < s_ip; i++)
		asA[0][i] = sprint("A", i+1);
	for(decl i = 0; i < s_iq; i++)
		asB[0][i] = sprint("B", i+1);	
	if(s_iDistribution == GAUSS)
		return {dOmega | vA | vB | dMu, {"omega"} ~ asA[0][] ~ asB[0][] ~ {"mu"} };
	else if(s_iDistribution == STUD_T)	
		return {dOmega | vA | vB | dMu | dDf, {"omega"} ~ asA[0][] ~ asB[0][] ~ {"mu", "df"} };
}

SystemMatrices(const dY, const dSigma2_t, const dDf, const avScore, const amInvFisher, const adllik)
{
	if(s_iDistribution == GAUSS){
		avScore[0] = (sqr(dY)-dSigma2_t)/(2*sqr(dSigma2_t));
		amInvFisher[0] = 2*sqr(dSigma2_t);
		adllik[0] = -log(dSigma2_t)/2 - sqr(dY)/(2*dSigma2_t);
	} else if(s_iDistribution == STUD_T){
		decl dw = (1 + (sqr(dY)/((dDf-2)*dSigma2_t)))^-1*((dDf+1)/(dDf-2));	
		avScore[0] = dw*sqr(dY)/(2*sqr(dSigma2_t)) - 1/(2*dSigma2_t);
		amInvFisher[0] = (2*sqr(dSigma2_t)*(dDf+3))/dDf;
		adllik[0] = log(gammafact((dDf+1)/2)) - log(gammafact(dDf/2)) - log((dDf-2)*M_PI)/2 - log(dSigma2_t)/2
				  - ((dDf+1)/2)*log(1+(sqr(dY)/((dDf-2)*dSigma2_t)));
	} else {
		println("Specify correct distribution type");
		return 0;
	}
	if(s_iLinkFunction == LOG_SIGMA){ 
		avScore[0] *= dSigma2_t; // Apply chain rule
		amInvFisher[0] /= sqr(dSigma2_t);
	}
	return 1;
}

Scaling(const amScaling, const mInvFisher)
{
	if(s_iScalingChoice == INV_FISHER){
		amScaling[0] = mInvFisher;
		return 1;
	} else if(s_iScalingChoice == INV_SQRT_FISHER){
		amScaling[0] = sqrt(mInvFisher);
		return 1;
	} else {
		println("Specify correct scaling type");
	}
	return 0;
}

Init(const dMu, const iMaxpq, const dDf)
{
	decl dsigma2_t, vscore, minvfisher, dllik, mscale;
	for(decl p = 0; p < iMaxpq; p++)
	{
		if(s_iLinkFunction == SIGMA)
			dsigma2_t = s_vf[p];
		else
			dsigma2_t = exp(s_vf[p]);
		SystemMatrices(s_mY[p]-dMu, dsigma2_t, dDf, &vscore, &minvfisher, &dllik);
		s_vScore[p] = vscore;
		Scaling(&mscale, minvfisher);
		s_vScaledSc[p] = mscale*vscore;
	}
}

TransformPar(const vP, const adomega, const avA, const avB, const admu, const addf)
{
	if(s_iLinkFunction == SIGMA)
		adomega[0] = exp(vP[0]);
	else
		adomega[0] = vP[0];
	avA[0] = reversec(vP[1:s_ip]);
	avB[0] = reversec(vP[1+s_ip:s_ip+s_iq]);
	if(s_iDistribution == GAUSS){
		admu[0] = vP[s_cPar-1];
		addf[0] = 0;
	}	
	else if(s_iDistribution == STUD_T){	
		admu[0] = vP[s_cPar-2];
		addf[0] = vP[s_cPar-1];
	}	
}

LogLikelihood(const vP, const adLlik, const avSco, const amHes)
{
	decl domega, vA, vB, dmu, imaxpq, dsigma2_t, vscore, minvfisher, mscale, iret, dllik, ddf, dloglik = 0;	
	TransformPar(vP, &domega, &vA, &vB, &dmu, &ddf);
	imaxpq = max(s_ip, s_iq);
	s_vf[:imaxpq-1] = double(domega/(1 - sumc(vB))); // Unconditional mean of factor f
	Init(dmu, imaxpq, ddf); // Init s_t	
	for(decl t = imaxpq; t < s_cT; t++)
	{				
		s_vf[t] = domega + s_vScaledSc[t-s_ip:t-1]*vA + s_vf[t-s_iq:t-1]*vB;
		if(s_iLinkFunction == SIGMA)
			dsigma2_t = s_vf[t];
		else
			dsigma2_t = exp(s_vf[t]); 
		iret = SystemMatrices(s_mY[t]-dmu, dsigma2_t, ddf, &vscore, &minvfisher, &dllik);
		if(s_iSandw == 1) s_vFunc[t-imaxpq] = dllik; // For sandwich standard errors		
		if(iret == 0) return 0;
		s_vScore[t] = vscore;
		dloglik += dllik;
		iret = Scaling(&mscale, minvfisher);
		if(iret == 0) return 0;
		s_vScaledSc[t] = mscale*vscore;
	}
	if(s_iDistribution == GAUSS) adLlik[0] = double((-log(M_2PI)*(s_cT-imaxpq)/2 + dloglik)/s_cT);
	else if(s_iDistribution == STUD_T) adLlik[0] = double(dloglik/s_cT);
	if(isnan(adLlik[0]) || sumc(vB) > 1) return 0;	
	return 1;   
}

ParStdErr(const vP)
{
    decl mhessian, minvhess, mvar, iresult;
	iresult = Num2Derivative(LogLikelihood, vP, &mhessian);		
	if(!iresult)
    {	print("Covar() failed in numerical second derivatives\n");
        return zeros(vP);
    }
	minvhess = invertgen(mhessian, 2); 
	if(s_iStdErr == HESS){
    	return sqrt(diagonal(-minvhess)/s_cT)';
	} else if(s_iStdErr == SAND){
		decl dh, mcontr_grad, dloglik, vllik, vptemp, vllik_h, vscore_t, minnerp;	 
		s_iSandw = 1;
		s_vFunc = zeros(s_cT-max(s_iq, s_ip), 1);
		dh = 1e-06;
		mcontr_grad = zeros(s_cT-max(s_iq, s_ip), s_cPar);
		LogLikelihood(vP, &dloglik, 0, 0);
		vllik = s_vFunc;
		for(decl i = 0; i < s_cPar; i++)
		{
			vptemp = vP;
			vptemp[i] += dh;
			LogLikelihood(vptemp, &dloglik, 0, 0);
			vllik_h = s_vFunc;
			mcontr_grad[][i] = (vllik_h-vllik)./dh; // Matrix of contributions to the gradient
		}
		minnerp = (mcontr_grad'*mcontr_grad)/s_cT;
		mvar = minvhess*minnerp*minvhess;
		s_iSandw = 0;
    	return sqrt(diagonal(mvar)/s_cT)';
	}
	println("Choose HESS or SAND to obtain standard errors");
}

MaxLik(vp, const aParNames)
{
    decl ir, dloglik, vpplot, vse;
    MaxControl(-1, 1, 1);
    ir = MaxBFGS(LogLikelihood, &vp, &dloglik, 0, TRUE);	
    println(MaxConvergenceMsg(ir), " using numerical derivatives");
	println("Log likelihood value = ", "%f", dloglik*s_cT);
	vse = ParStdErr(vp);
	vpplot = vp;
	if(s_iLinkFunction == SIGMA){
		vse[0] = vse[0]*exp(vp[0]); // Delta method
		vpplot[0] = exp(vp[0]);
	}
	print("Parameters with standard errors:", "%cf", {"%12.5g", " (%7.5f)"}, "%r", aParNames, vpplot ~ vse);
	return vp;
}

PlotData()
{
	//DrawAxis(0, 0, 1980, -0.2, 0.15, -0.1, 0.1, 0.02, 0);
	DrawTMatrix(0, s_mY, "Continuously compounded return", 1980, 1, 53, 0, 3);
	if(s_iLinkFunction == SIGMA)
		DrawTMatrix(0, sqrt(s_vf), "Estimated volatility (sigma_t)", 1980, 1, 53, 0, 2);
	else if(s_iLinkFunction == LOG_SIGMA)
		DrawTMatrix(0, sqrt(exp(s_vf)), "Estimated volatility (sigma_t)", 1980, 1, 53, 0, 2);
	//DrawAxis(1, 0, 1980, -5e3, 4e4, 0, 2e4, 4e3, 0);
	DrawTMatrix(1, s_vScore, "Score", 1980, 1, 53, 0, 3);
	DrawTMatrix(2, s_vScaledSc, "Scaled score", 1980, 1, 53, 0, 3);	
}

CompareGaussStudt(const domega, const vA, const vB, const dmu, const ddf)
{
	decl vp, vfgauss, vfstudt, aparnames;
	s_iDistribution = GAUSS;
	[vp, aparnames] = StartingValues(domega, vA, vB, dmu, ddf);
	s_cPar = rows(vp);
	vp = MaxLik(vp, aparnames);
	vfgauss = s_vf;
	s_iDistribution = STUD_T;
	[vp, aparnames] = StartingValues(domega, vA, vB, dmu, ddf);
	s_cPar = rows(vp);
	vp = MaxLik(vp, aparnames);
	vfstudt = s_vf;
	DrawAxis(0, 0, 1980, -0.2, 0.2, -0.2, 0.05, 0, 0);
	DrawTMatrix(0, s_mY, "Continuously compounded return", 1980, 1, 53, 0, 1);
	if(s_iLinkFunction == SIGMA){
		DrawTMatrix(0, sqrt(vfgauss), "Estimated volatility (sigma_t) gauss", 1980, 1, 53, 0, 2);
		DrawTMatrix(0, sqrt(vfstudt), "Estimated volatility (sigma_t) stud_t", 1980, 1, 53, 0, 3);
	}	
	else if(s_iLinkFunction == LOG_SIGMA){
		DrawTMatrix(0, sqrt(exp(vfgauss)), "Estimated volatility (sigma_t) gauss", 1980, 1, 53, 0, 2);
		DrawTMatrix(0, sqrt(exp(vfstudt)), "Estimated volatility (sigma_t) stud_t", 1980, 1, 53, 0, 3);
	}	
	ShowDrawWindow();	
}

main()
{
	// Univariate GAS model for volatility. See GAScode.pdf for explanation of the program
	// Author: Rutger Lit
	decl dtime = 0, mdata, dscaling, dloglik, domega, vA, vB, dmu, ddf, vp, aparnames;

//*************************************************************************************************
//  START OF USER INPUT	
//	Load data										   
//	Specify scaling, distribution, link function, order of gas model,	   
//	choice of standard errors and starting values 					   
//*************************************************************************************************

	// Load data
	mdata = loadmat("DJInd19801999.xls"); s_mY = mdata[1:][4]';
	dscaling = 1; // Scaling data can improve stability, 1 for no scaling
	s_mY.*= dscaling;
	// Distribution: GAUSS, STUD_T
	s_iDistribution = GAUSS;
	// Link function: SIGMA (f_t=sigma^2_t), LOG_SIGMA (f_t=log(sigma^2_t))
	s_iLinkFunction = LOG_SIGMA;
	// Scaling score: INV_FISHER, INV_SQRT_FISHER
	s_iScalingChoice = INV_FISHER;
	// Order of GAS model p, q
	s_ip = 1; s_iq = 1;
	// Standard erros: HESS, SAND
	s_iStdErr = HESS;	
	// Starting values (note the dimensions of s_ip and s_iq)
	domega = 0;  
	vA = <0.10>'; // Extend for higher orders of p, separate with comma's. 
	vB = <0.89>'; // Extend for higher orders of q, separate with comma's. B_1 + ... + B_q < 1
	dmu = 0;
	ddf = 5; // Only estimated if s_iDistribution = STUD_T
	
//*************************************************************************************************
//  END OF USER INPUT							      
//*************************************************************************************************

	s_cT = columns(s_mY);
	s_vf = s_vScore = s_vScaledSc = zeros(1, s_cT);
	[vp, aparnames] = StartingValues(domega, vA, vB, dmu, ddf); // Note that the number of elements in vp varies with the choice of distribution
	s_cPar = rows(vp);
	//LogLikelihood(vp, &dloglik, 0, 0);
	//println("Log Likelihood value = ", "%f", dloglik*s_cT);
	vp = MaxLik(vp, aparnames);
	PlotData();
	ShowDrawWindow();	
	//CompareGaussStudt(domega, vA, vB, dmu, ddf);
	println("Time: ", timespan(dtime));	
}
